/**
 * =============================================================================
 *  Name        : fluid.cl
 *  Part of     : oclcomp
 *  Description : OpenCL .cl file
 *  Version     : 0.09.06.30
 *
 *  Copyright (c) 2008-2009 Nokia.  All rights reserved.
 *  This material, including documentation and any related computer
 *  programs, is protected by copyright controlled by Nokia.  All
 *  rights are reserved.  Copying, including reproducing, storing,
 *  adapting or translating, any or all of this material requires the
 *  prior written consent of Nokia.  This material also contains
 *  confidential information which may not be disclosed to others
 *  without the prior written consent of Nokia.
 *
 * =============================================================================
**/


#include "real.h"


__kernel
void diffuse( _real a, _real b, _real* d, const _real* d0, const int w, const int h)
{
	const int c = (get_global_id(1) + 1) * w + get_global_id(0) + 1;
	d[c] = mult((d0[c] + mult(a, d[c-w]+d[c+w]+d[c-1]+d[c+1])), b);
}


__kernel
void advect(_real *d, _real *d0, _real *u, _real *v, _real dtw, _real dth, const int w, const int h)
{
	int i0, j0, i1, j1;
	_real x, y, s0, t0, s1, t1;
	const int j = get_global_id(0)+1;
	const int i = get_global_id(1)+1;
	const int p = i*w+j;
	int pi0, pi1;
	y = int2real(i) - mult(dth, u[p]);
	x = int2real(j) - mult(dtw, v[p]);
	x = x < F_HALF? F_HALF: x > int2real(w) - F_ONEHALF? int2real(w) - F_ONEHALF: x;
	//printf("x=%d y=%d\n", real2int(x), real2int(y));
	j0 = real2int(x);
	j1 = j0 + 1;
	s1 = x - int2real(j0);
	s0 = F_ONE - s1;
	y = y < F_HALF? F_HALF: y > int2real(h) - F_ONEHALF? int2real(h) - F_ONEHALF: y;
	i0 = real2int(y);
	i1 = i0 + 1;
	t1 = y - int2real(i0);
	t0 = F_ONE - t1;
	pi0 = i0 * w;
	pi1 = i1 * w;
	//printf("x=%d y=%d, ", real2int(x), real2int(y));
	//printf("i0=%d j0=%d i1=%d j1=%d\n", i0, j0, i1, j1);
	d[p] = mult(t0, mult(s0, d0[pi0+j0]) + mult(s1, d0[pi0+j1])) +
		   mult(t1, mult(s0, d0[pi1+j0]) + mult(s1, d0[pi1+j1]));
}


__kernel
void project0(const _real *u, const _real *v, _real *p, _real *div, const _real hw, const _real hh, const int w, const int h)
{
	const int c = (get_global_id(1) + 1) * w + get_global_id(0) + 1;
	div[c] = mult(hw, (u[c+w]-u[c-w])) + mult(hh, (v[c+1]-v[c-1]));
	p[c] = 0;
}


__kernel
void project1(_real *p, const _real *div, const int w, const int h)//, _real *test) // remove test, was p!
{
	const int c = (get_global_id(1) + 1) * w + get_global_id(0) + 1;
	p[c] = mult( div[c]+p[c-w]+p[c+w]+p[c-1]+p[c+1], F_ONE / 4);
}


__kernel
void project2(_real *u, _real *v, const _real *p, const _real nw2, const _real nh2, const int w, const int h)
{
	const int c = (get_global_id(1) + 1) * w + get_global_id(0) + 1;
	u[c] -= mult((p[c+w]-p[c-w]), nw2);
	v[c] -= mult((p[c+1]-p[c-1]), nh2);
}


__kernel
void buoyancy(_real *u, _real *v, const _real *d, const _real uforce, const _real vforce, const _real ave, const int w)
{
	const int c = (get_global_id(1) + 1) * w + get_global_id(0) + 1;
	_real dif = ave - d[c];
    u[c] += mult(dif, uforce);
    v[c] += mult(dif, vforce);
}

