/**
 * =============================================================================
 *  Name        : trysph.cl
 *  Part of     : oclcomp
 *  Description : OpenCL .cl file
 *  Version     : 0.09.06.30
 *
 *  Copyright (c) 2008-2009 Nokia.  All rights reserved.
 *  This material, including documentation and any related computer
 *  programs, is protected by copyright controlled by Nokia.  All
 *  rights are reserved.  Copying, including reproducing, storing,
 *  adapting or translating, any or all of this material requires the
 *  prior written consent of Nokia.  This material also contains
 *  confidential information which may not be disclosed to others
 *  without the prior written consent of Nokia.
 *
 * =============================================================================
**/


#include <stdio.h>
#include "trysph_def.h"


__kernel
void relaxation(float2 *Pos, int *Grid, float *Pres, _float Touch)
{
	_float xi, yi, xj, yj, dx, dy, d2, d, df, d_rec, mx, my, dfx, dfy;
	_float dens, dens_near, a, aa;
	_float pres, pres_near;
	int neighbors[MAX_PARTICLES];
	_float ds[MAX_PARTICLES];
	_float as[MAX_PARTICLES];
	int i, j, n, k;
    int gx, gy, gj, gi;
    int phase = get_global_id(2);
    int gxb = get_global_id(0)*Grid_parallel_step + ((phase&1)*Grid_parallel_step/2);
    int gyb = get_global_id(1)*Grid_parallel_step + ((phase&2)/2*Grid_parallel_step/2);

    for (gy = gyb; gy < gyb + Grid_parallel_step/2; gy++) {
      for (gx = gxb; gx < gxb + Grid_parallel_step/2; gx++) {
        i = Grid[gy*Grid_w+gx];
        if (i==-1) continue;
	    xi = Pos[i].x;
	    yi = Pos[i].y;
        // note! our position gy,gx may not be absolutely correct if shifted due to high occupacy
        //gx = get_grid_x(xi);
        //gy = get_grid_y(yi);
        if (i == 0) {
            if (Touch <= 0.01) continue; // don't consider particle 0 when no touch active
            dens = Touch;
        }
		else dens = 0;
		dens_near = 0;
        // TEST: GET BASIC DENSITY BASED ON POSITION - could be e.g. image brightness based
        //_float xid=xi-.5, yid=yi-.5;
        //dens = xid*xid+yid*yid*Density*10;
        //dens_near = dens;
		n = 0;
		for (gj = gy-Grid_sph_r; gj <= gy+Grid_sph_r; gj++)
		{
            // NOTE! It could be possible to make chamber edges to look like particles here
            // => better chambering. I.e. each out-of-bounds neighbor cell would add density.
            if (gj<0 || gj>=Grid_h) continue;
		    for (gi = gx-Grid_sph_r; gi <= gx+Grid_sph_r; gi++)
		    {
                if (gi<0 || gi>=Grid_w) continue;
                j = Grid[gj*Grid_w+gi];
                if (j==-1) continue;
                if (j==0 && Touch <= 0.01) continue; // don't consider particle 0 when no touch active
			    if (j != i)
			    {
				    xj = Pos[j].x;
				    yj = Pos[j].y;
				    dx = xi - xj;
				    dy = yi - yj;
				    d2 = dx * dx + dy * dy;
				    if (d2 < Sph_radius2)
				    {
					    neighbors[n] = j;
					    d = sqrt(d2);
					    ds[n] = d;
					    a = 1.0 - d / Sph_radius;
					    as[n] = a;
					    aa = a * a;
					    dens += aa;
					    dens_near += aa * a;
					    n++;
				    }
			    }
		    }
        }        
		pres = Stiffness * (dens - Density);
        Pres[i] = pres;
		//pres_near = Stiffness_near * (dens_near - Density);
		pres_near = Stiffness_near * dens_near;
		//if (i==0)
			//printf("dens=%5f dens_near=%5f pres=%5f pres_near=%5f\n",
				//dens, dens_near, pres, pres_near);
		mx = 0;
		my = 0;
		for (k = 0; k < n; k++)
		{
			j = neighbors[k];
			xj = Pos[j].x;
			yj = Pos[j].y;
			if (ds[k] != 0.0) { // check if particles are on top of each other
				d_rec = -0.5 / ds[k]; // halved here, already
				dx = d_rec * (xi - xj);
				dy = d_rec * (yi - yj);
				a = as[k];
                // NOTE! How about partial relaxation here, to get fluid more compressible
                // to enable a kind of 3D-effect. <- tested, didn't gave good results.
				df = pres * a + pres_near * a * a;
				df *= Timestep * Timestep;
				dfx = df * dx;
				dfy = df * dy;
				Pos[j].x += dfx;
				Pos[j].y += dfy;
				mx -= dfx;
				my -= dfy;
			}
			//else printf("warning: singularity\n");
		}
		Pos[i].x += mx;
		Pos[i].y += my;
		//if (i==0) printf("dist=%f\n", Dist[i]);
    }
  }
}

